/*
*	trans: string translator
*	Brute Force
*	ver. 1.0
*	-Kenar- (Kenji Arisawa)
*	Email: arisawa@aichi-u.ac.jp
*/

/*
*	PLAN9:	Plan9 and P9P
*	UNIX:	UNIX and P9P
*/

#ifdef PLAN9
/* include Plan9 port */
#include <u.h>
#include <libc.h>
#include <bio.h>
#else
#include <plan9.h>
#include <stdlib.h>	/* for malloc */
#include "ubio.h"
/*	print() (in print.c) is buggy, so I replaced by printf() in stdio.h .
* look also "Segmentation fault" in this file. */
#define print printf
#define UNIX	1
int open(char*,int);
#endif

#include "trans.h"

#define	DBG	if(0)
#define	DBG0 if(0)
/*	note: if(0) { ... }
*	if(0) strips the object codes from *.$O.
*	you need not remove debugging codes from the source to save object code size.
*/

#ifdef UNIX
#define stime s_time
#endif

#ifndef PLAN9

#endif



Biobuf bout;

char *
etrans(char *result, char *ep, char *source)
/*	Brute Force
*	result: buf to store the result
*	ep: size of result
*	source: source text
*	ttab: table of rule
*	n: number of rules
*	return: next writing point (not null terminated)	*/
{
	int m,j,n;
	char *s;
	char *wp = result;

	n = nrule;
	DBG print("n:%d\n",n);

	for(s = source; *s;s++){
		for(j = n - 1; j >= 0; j--){
			for(m = 0; s[m] && m < ttab[j].frlen; m++)
				if(s[m] != ttab[j].frstr[m])
					break;
			if(m == ttab[j].frlen){ /* matched */
				s += m - 1; /* NB: next s++ */
				break;
			}
		}
		if(j < 0)
			wp = ewrite(wp, ep, s, 1);
		else
			wp = ewrite(wp, ep, ttab[j].tostr, strlen(ttab[j].tostr));
	}
	return wp;
}

/*
*	tabread: read translation data from rule file
*	lines beginning with "#" are comment
*	if the first line is comment and string "unicode" or "hex" is in it,
*	these are keywords for format of the file 	*/
void
tabread(char *path)
{
	int n,m;
	Trtab trtab;
	n = readtable0(&trtab,path,0);
	ttab = trtab.ttab;
	m = trtab.max;
	/* then m is max size of frlen */
	trbuf = malloc(m + 1); // need space for '\0' in case etrans()
	nrule = n;
}


/*
*	stranslate() shows the usage example of etrans()
*	etrans() is designed for string to string conversion
*	these string must be utf-8.
*/
void
stranslate(int fd)
{
	Biobuf bin;
	char *source;
	char obuf[4098], *ep, *wp;
	ep = obuf + sizeof(obuf);
	if(Binit(&bin, fd, OREAD) < 0)
		sysfatal("input not open: %r");
	while((source = Brdstr(&bin,'\n',1))){ /* assign = */
		DBG0 print("!%s\n",source);
		wp = etrans(obuf, ep, source);
		if(wp == ep)
			fprint(2,"# output buffer full\n");
		*wp = 0; /* needed */
		Bprint(&bout,"%s\n",obuf);
	}
	Bflush(&bout);
}

/* 	we have two types of translate()
*	Btranslate() and stranslate()
*	Btranslate() is better for the tool that handle stream
*	stranslate() is designed for string conversion.
*	stranslate() is only for debugging purpose.
*	we need the codes in stranslate() for u9fs */



#define translate stranslate

void
usage(void)
{
	fprint(2,
	"usage: trans rulefile [file ....]\n"
	"name:\n"
	"  translate strings using rules in the rulefile\n"
	);
	exits("usage");
}

void
main(int argc, char *argv[])
{
	int fd;

	ARGBEGIN{
	default: usage();
	}ARGEND

	if(*argv == nil)
		usage();

	Binit(&bout, 1, OWRITE);

	tabread(*argv++);

	if(*argv == nil){
		translate(0);
		exits(nil);
	}

	while(*argv){
		fd = open(*argv++,OREAD);
		translate(fd);
	}
	exits(nil);
}

