/*
*	translate strings
*	Trie-Based Algorithm
*	loose implementation of  Aho-Corasick
*
*	NB: Trie-Based Algorithm is beautiful but waste too much memory.
*/



#include <u.h>
#include <libc.h>
#include <bio.h>
#include "trans.h"

#define	DBG	if(0)
#define	DBG0	if(0)

/*	Trie: retrieval tree
*	Trie enables fast algorithm but wastes much memory	*/
typedef struct Trie Trie;
struct Trie {
char *frstr;
char *tostr;
Trie *node[256];
} *troot;

Biobuf bout;

void
usage(void)
{
	fprint(2,
	"usage: trans rulefile [file ....]\n"
	);
	exits("usage");
}


void
Tput(char *frstr, char *tostr)
{
	Trie *t = troot;
	uchar *s;
	for(s = (uchar*)frstr; *s; s++){
		if(t == nil)
			t = calloc(1, sizeof(Trie));
		if(t->node[*s] == nil)
			t->node[*s] = calloc(1, sizeof(Trie));
		t = t->node[*s];
	}
	t->frstr = strdup(frstr);
	t->tostr = strdup(tostr);
}

/*
*	tabread: read translation data from rule file
*	lines beginning with "#" are comment
*	if the first line is comment and string "unicode" or "hex" is in it,
*	these are keywords for format of the file 	*/
void
tabread(char *path)
{
	int n,m, fd;
	int fmt = 0;
	Biobuf bin;
	char *p;
	char *args[3];
	long ch;

	/*
	*	trans data init
	*/
	fd = open(path, OREAD);
	if(fd < 0)
		sysfatal("%s not open %r", path);
	if(Binit(&bin, fd, OREAD) < 0)
		sysfatal("error in Binit: %r");
	ch = Bgetc(&bin);
	if(ch == '#'){
		p = Brdstr(&bin, '\n',1);
		if(cistrstr(p,"unicode"))
			fmt = ufmt;
		else
		if(cistrstr(p,"hex"))
			fmt = hfmt;
	}
	else
		Bungetc(&bin);

	n = 0;
	while((p = Brdstr(&bin, '\n', 1))){ /* assign = */
		if(p[0] == '#')
			continue;
		m = tokenize(p,args, 3);
		if(m < 2)
			continue;
		Tput(trfmt(args[0],fmt),trfmt(args[1],fmt));
		n++;
	}
	close(fd);
	if(n == 0)
		sysfatal("# no rule");
}

#define OUTPUT(s,n) Bwrite(&bout,s,n)

void
trans(char *text)
/*	source: source text
*	ttab: table of rule
*	n: number of rules
*	buf: working buf, the size must be >= max len of ttab[n].frstr	*/
{
	uchar *s;
	char *t,*u;
	uchar buf[256];
	int m;
	Trie *d;

	for(s = (uchar*)text; *s; s++){
		d = troot;
		t = (char*)s;
		while(*s && d->node[*s] == nil){
			s++;
		}
		OUTPUT(t, (char*)s - t);
		t = (char*)s;
		m = 0;
		while(*s && d->node[*s]){
			d = d->node[*s];
			buf[m++] = *s;
			s++;
		}
		if(*s == 0){
			u = d->frstr;
			if(u && strcmp(t,u) == 0){
				DBG print("FOUND: %s ; %s ; %s\n", t, d->frstr, d->tostr);
				OUTPUT(d->tostr,strlen(d->tostr));
			}
			else
				OUTPUT(t,strlen(t));
			break;
		}
		if(d->node[*s] == nil){
			if(d->tostr){
				DBG print("FOUND: %s ; %s ; %s\n", t, d->frstr, d->tostr);
				OUTPUT(d->tostr,strlen(d->tostr));
				s--;
				continue;
			}
		}
		OUTPUT((char*)buf,1);
		s -= m;
	}
}


/*
*	stranslate() shows the usage example of etrans()
*	etrans() is designed for string to string conversion
*	these string must be utf-8.
*/
void
translate(int fd)
{
	Biobuf bin;
	char *source;
	if(Binit(&bin, fd, OREAD) < 0)
		sysfatal("input not open: %r");
	while((source = Brdstr(&bin,'\n',1))){ /* assign = */
		trans(source);
		Bwrite(&bout,"\n",1);
	}
	Bflush(&bout);
}


void
main(int argc, char *argv[])
{
	int fd;
	ARGBEGIN{
	default: usage();
	}ARGEND

	if(*argv == nil)
		usage();

	Binit(&bout, 1, OWRITE);

	troot = calloc(1, sizeof(Trie));

	tabread(*argv++);
	DBG print("# tabread done\n");

	if(*argv == nil){
		translate(0);
		exits(nil);
	}

	while(*argv){
		fd = open(*argv++,OREAD);
		translate(fd);
	}

	exits(nil);
}
