/*
*	trans: string translator
*	Sandwich Method by Kenar
*	ver. 1.0
*	-Kenar- (Kenji Arisawa)
*	Email: arisawa@aichi-u.ac.jp
*/

/*
*	PLAN9:	Plan9 and P9P
*	UNIX:	UNIX and P9P
*/

#ifdef PLAN9
/* include Plan9 port */
#include <u.h>
#include <libc.h>
#include <bio.h>
#else
#include <plan9.h>
#include <stdlib.h>	/* for malloc */
#include "ubio.h"
/*	print() (in print.c) is buggy, so I replaced by printf() in stdio.h .
* look also "Segmentation fault" in this file. */
#define print printf
#define UNIX	1
int open(char*,int);
#endif

#include "trans.h"

#define	DBG	if(0)
#define	DBG0 if(0)
/*	note: if(0) { ... }
*	if(0) strips the object codes from *.$O.
*	you need not remove debugging codes from the source to save object code size.
*/

#ifdef UNIX
#define stime s_time
#endif

List *lroot;
Ttab *ttab;
char *trbuf;
int nrule;
Biobuf bout;
int binmode;

/*
*	u8rb:	gets the second utf-8 rune boundary
*	note that the first rune boundary is &bp[0] */
int
u8rb(char *bp, int m)
{
	char *p,*e;
	p = bp;
	e = bp + m;
	p++;
	while(p < e){
		if((*p & 0x80) == 0)
			break;
		if((*p & 0xc0) == 0xc0)
			break;
		p++;
	}
	return p - bp;
}



void
sstrans(Biobuf *bop, Biobuf *bip)
/*	bop: pointer to bout
*	bip: pointer to bin
*	ttab: table of rule
*	n: number of rules
*	buf: working buf, the size must be >= max len of ttab[n].frstr	*/
{
	int m,j,k,c;
	int n;
	char *buf;
	int i;

	buf = trbuf;
	n = nrule;
	DBG print("n:%d\n",n);

	for(;;){
Lstart:
		m = 0; j = 0; k = n - 1;
		while(j < k) {
			c = Bgetch(bip);
			DBG print("|%02x:m,j,k:%d,%d,%d\n",c,m,j,k);
			if(c == ERR)
				goto Lend;
			while (j <= k && c > (uchar)ttab[j].frstr[m]) j++;
			while (j <= k && c < (uchar)ttab[k].frstr[m]) k--;
			buf[m++] = c;
			DBG print(">%02x:m,j,k:%d,%d,%d\n",c,m,j,k);
		}
		DBG print("*%02x:m,j,k:%d,%d,%d\n", c,m,j,k);
		if(j == k) {
			while(m < ttab[j].frlen) {
				c = Bgetch(bip);
				DBG print("^%02x,%02x:m,frlen:%d,%d\n",c,(uchar)ttab[j].frstr[m],m,ttab[j].frlen);
				if(c == ERR)
					goto Lend;
				buf[m] = c;
				/* NB: m is number of bytes in buffer */
				if((uchar)ttab[j].frstr[m++] != c)
					goto Lnext;
			}
			DBG print("@%02x:m,j:%d,%d\n",c,m,j);
			if(m == ttab[j].frlen) /* found */
				Bwrite(bop, ttab[j].tostr, strlen(ttab[j].tostr));
			goto Lstart;
		}
Lnext: /* flash buf and go next */
		if(binmode)
			i = 1;
		else
			i = u8rb(buf, m);
		Bwrite(bop, buf, i);
		Bungetch(bip, &buf[i], m - i);
	}
Lend:
	Bwrite(bop, buf, m);
}


char *
mmtrans(char *result, char *ep, char *source)
/*	result: buf to store the result
*	ep: size of result
*	source: source text
*	ttab: table of rule
*	n: number of rules
*	buf: working buf, the size must be >= max len of ttab[n].frstr
*	return: next writing point (not null terminated)	*/
{
	int m,j,k,c;
	int n;
	char *buf, *s;
	int i;
	char *wp = result;

	buf = trbuf;
	n = nrule;
	DBG print("n:%d\n",n);
	s = source;
	for(;;){
Lstart:
		m = 0; j = 0; k = n - 1;
		while(j < k) {
			c = *s++;
			if(c == 0)
				goto Lend;
			DBG print("|%02x:m,j,k:%d,%d,%d\n",c,m,j,k);
			while (j <= k && c > (uchar)ttab[j].frstr[m]) j++;
			while (j <= k && c < (uchar)ttab[k].frstr[m]) k--;
			buf[m++] = c;
			DBG print(">%02x:m,j,k:%d,%d,%d\n",c,m,j,k);
		}
		DBG print("*%02x:m,j,k:%d,%d,%d\n", c,m,j,k);
		if(j == k) {
			while(m < ttab[j].frlen) {
				c = *s++;
				DBG print("^%02x,%02x:m,frlen:%d,%d\n",c,(uchar)ttab[j].frstr[m],m,ttab[j].frlen);
				if(c == 0)
					goto Lend;
				buf[m] = c;
				/* NB: m is number of bytes in buffer */
				if((uchar)ttab[j].frstr[m++] != c)
					goto Lnext;
			}
			DBG print("@%02x:m,j:%d,%d\n",c,m,j);
			if(m == ttab[j].frlen) /* found */
				wp = ewrite(wp,ep,ttab[j].tostr, strlen(ttab[j].tostr));
			goto Lstart;
		}
Lnext: /* flash buf and go next */
		if(binmode)
			i = 1;
		else
			i = u8rb(buf, m);
		wp = ewrite(wp, ep, buf, i);
		s -= m - 1;
	}
Lend:
	wp = ewrite(wp,ep, buf, m);
	return wp;
}

/*
*	tabread: read translation data from rule file
*	lines beginning with "#" are comment
*	if the first line is comment and string "unicode" or "hex" is in it,
*	these are keywords for format of the file 	*/
void
tabread(char *path)
{
	int n,m;
	Trtab trtab;
	n = readtable0(&trtab,path,1);
	ttab = trtab.ttab;
	m = trtab.max;
	/* then m is max size of frlen */
	trbuf = malloc(m + 1); // need space for '\0' in case mmtrans()
	nrule = n;
}


/*
*	sstranslate() shows the usage example of sstrans()
*	sstrans() is designed for stream to stream conversion
*	stdin to stdout
*	Binary data should be OK.
*/
void
sstranslate(int fd)
{
	Biobuf bin;
	if(Binit(&bin, fd, OREAD) < 0)
		sysfatal("input not open: %r");
	sstrans(&bout, &bin);
	Bflush(&bout);
}


/*
*	mmtranslate() shows the usage example of mmtrans()
*	mmtrans() is designed for memory to memory conversion
*	input must be utf-8.
*/
void
mmtranslate(int fd)
{
	Biobuf bin;
	char *source;
	char obuf[4098], *ep, *wp;
	ep = obuf + sizeof(obuf);
	if(Binit(&bin, fd, OREAD) < 0)
		sysfatal("input not open: %r");
	while((source = Brdstr(&bin,'\n',1))){ /* assign = */
		DBG0 print("!%s\n",source);
		wp = mmtrans(obuf, ep, source);
		if(wp == ep)
			fprint(2,"# output buffer full\n");
		*wp = 0; /* needed */
		Bprint(&bout,"%s\n",obuf);
	}
	Bflush(&bout);
}


void
usage(void)
{
	fprint(2,
	"usage: trans [-b] rulefile [file ....]\n"
	"name:\n"
	"  translate strings using rules in the rulefile\n"
	"options:\n"
	" -b: binary mode\n"
	);
	exits("usage");
}


/* 	we have two types of translate()
*	Btranslate() and stranslate()
*	Btranslate() is better for the tool that handle stream
*	stranslate() is designed for string conversion.
*	stranslate() is only for debugging purpose.
*	we need the codes in stranslate() for u9fs */

#define translate sstranslate

void
main(int argc, char *argv[])
{
	int fd;

	ARGBEGIN{
	default: usage();
	}ARGEND

	if(*argv == nil)
		usage();

	Binit(&bout, 1, OWRITE);

	tabread(*argv++);

	if(*argv == nil){
		translate(0);
		exits(nil);
	}

	while(*argv){
		fd = open(*argv++,OREAD);
		translate(fd);
	}
	exits(nil);
}
